"""
Copyright 2019 AIST

Licensed under the Apache License, Version 2.0 (the “License”);
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an “AS IS” BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


# The second. Read a XML sample, and then write it out to a file.
import os
import sys
import getopt
import xml.etree.ElementTree as etree
import xml.dom.minidom as miniDom
import json
import time
import re
import tkinter
import tkinter.messagebox as tkmsg


#
#	global veriables
#
lgfd = None		# file descriptor for the log file
debug = False		# command line option
comp = {}		# <components> part of the source XML
libs = {}		# <libparts> part of the source XML
nets = {}		# <nets> part of the source XML
devices = {}
pseudodevices = {}
urnprifix = None
userFields = {"GLPK","GLPKtype","GLPKchannelTableId","pinAttribute","Model","Type","Controller","Socket","LocalConfig","LogicalName","Cost"}
costEquation = None	#  コスト評価式
unitLoss = {}		# 単位長当たりの損失のデフォルト値をファイバの種類ごとに覚えておく。キーはファイバの種類で値は実数
unitDispersion = {}	# 分散についても同様
channelInfo = {}	# 一旦読み込んでそのまま出力する
signalformInfo = {}
maxLength = 0.0
maxLoss = 0.0
maxDispersion = 0.0


#
#	msglog() ----- エラーなどのメッセージの記録または表示。quitがエラーのレベルを表し、以下のような動作となる：
#	quit	意味	動作
#	0	info	debugフラグが立っていればメッセージ・ダイアログをポップアップ
#	1	warring	メッセージ・ダイアログをポップアップして続行するか問う
#	≧2	error	メッセージ・ダイアログをポップアップしてOK後終了
#	
#	ログファイルの生成が指定されていればすべてのレベルでメッセージを記録する。

def msglog(msg, quit = 0):
	'''When the log-file is opened, ``msg'' will be written into the log-file.
	Otherwisem it will be sent to stdout. And the execution is terminated
	if the second argunent is > 0.
	'''
	if lgfd != None:
		lgfd.write(msg+"\n")
	if quit == 0 and debug:
		tkmsg.showinfo("xml2topology", msg)

	if quit > 1:
		tkmsg.showerror("xml2topology", msg)
		if lgfd != None:
			lgfd.write("Further execution terminated.\n")
			lgfd.close()
		sys.exit(quit)
	else:	# quit == 1
		if not tkmsg.askyesno("xml2topology", msg+"\ncontinue?"):
			if lgfd != None:
				lgfd.write("Further execution terminated.\n")
				lgfd.close()
			sys.exit(quit)


#
#	extraAttr() ----- looks for a net that connected to PORTATTR type device.
#		When it is found, copy attribute of the device into the port.
#

def extraAttr(refnum, pinnum, nets, comp):
	errcnt = 0
	for np in nets:
		if np.tag == "net":
			for rp in np:
				if rp.tag == "node":
					ref = rp.attrib['ref']
					pin = int(rp.attrib['pin'])
					if ref == refnum and pin == pinnum: # find a net that is connected to the port.
						# Look for the opposite side in component list
						for op in np:
							if op != rp:
								attrref = op.attrib['ref']
								for cp in comp:
									if cp.tag == 'comp' :
										if cp.attrib['ref'] == attrref:
											flds = cp.find("fields")
											if flds == None:
												return None
											portattrs = {}
											attr = False
											for fp in flds:
												if fp.tag == 'field'and "name" in fp.attrib:
													if fp.attrib['name'] == "Type" and fp.text == "PORTATTR":
														attr = True
													elif fp.text != "":
														portattrs[fp.attrib['name']] = fp.text
											if attr:
												return portattrs
											else:
												return None
						return None
	return None


#
#	findNetByRefPin() ----- look for a net that has specified ref and pin in nets.
#

def findNetByRefPin(nets, sref, spin):
	for np in nets:
		if np.tag == "net":
			for sp in np:
				if sp.tag == "node":
					if sref == sp.attrib["ref"] and spin == sp.attrib["pin"]:
						return(np)
	return(None)


#
#	isConnectToPseudoDevice() ----- 指定されたネット内で、ref及びpinと反対側のデバイスが仮想デバイスか否かを調べる。
#

def isConnectToPseudoDevice(comp, net, ref, pin):
	for np in net:
		if np.tag == "node":
			if np.attrib["ref"] != ref or np.attrib["pin"] != pin:
				for cp in comp:
					if cp.attrib["ref"] == np.attrib["ref"]:
						flds = cp.find("fields")
						for fp in flds:
							if fp.attrib['name'] == "Type" and fp.text == "PORTATTR":
								return(True)
	return(False)


#
#	getFarSide() ----- 指定されたnet内でrefとpinでない方のnodeを返す
#

def getFarSide(net, ref, pin):
	for sp in net:
		if sp.tag == "node":
			if ref != sp.attrib["ref"] or pin != sp.attrib["pin"]:
				return(sp)
	return(None)


#
#	markRemove() ----- replace tag "net" to "removed"
#

def markRemove(nets, nn):
	for np in nets:
		if np.tag == "net":
			if np.attrib["code"] == nn.attrib["code"] and np.attrib["name"] == nn.attrib["name"]:
				np.tag = "removed"
				break
	return(None)

#
#	componentXml() ----- reconstruct XML for device list
#

def componentXml(devices):
	cxml = etree.Element("components")
	for ref in devices:
		cx = etree.Element("comp")
		cx.set("ref", ref)	# ref# is given as an attribute
		gp = None
		for itemkey in devices[ref]:
			if itemkey != "ports":	# all other items are coded as "field"
				if not itemkey in ["GLPKtype","GLPKchannelTableId"]:	# These two fields will be put into GLPK field as its attribute.
					ix = etree.Element("field")
					ix.set("name", itemkey)
					ix.text = devices[ref][itemkey]
					cx.append(ix)
					if itemkey == "GLPK":
						gp = ix
			else:			# ports
				px = etree.Element("ports")
				for pp in devices[ref][itemkey]:
					qx = etree.Element("port")
					qx.set("number", pp)
					for ap in devices[ref][itemkey][pp]:
						qx.set(ap, devices[ref][itemkey][pp][ap])
					px.append(qx)
				cx.append(px)
		if gp != None:
			for itemkey in ["GLPKtype","GLPKchannelTableId"]:
				if itemkey in devices[ref]:
					gp.set(itemkey, devices[ref][itemkey])	# given as an attribute
		cxml.append(cx)
	return(cxml)


#
#	setPair() ----- 属性"pair"が既にあれば、ペアの相方の名前を追加する。なければ生成して相方の名前を登録する。
#

def setPair(node, name):
	pp = node.get("pair", default=None)
	if pp == None:
		node.set("pair", name)
	elif pp.find(name) < 0:		# 二重登録を防ぐ
		node.set("pair", pp+","+name)


#
#	pairPath() ----- 自動生成ではない名前（Net-...でない名前）の場合、先頭からが同一で末尾が-[01]である配線はペアと見做して印をつける。
#	pairPath() ----- 自動生成ではない名前（Net-...でない名前）の場合、先頭からが同一で末尾が-[0-9]+である配線はペアと見做して印をつける。
#

def pairPath(nets):
	for np in nets:
		if np.tag != "net":
			continue
		name = np.attrib["name"]
		if re.match(r"^Net-\([A-Za-z0-9\-]+\)", name) != None:	# a name given automatically by KiCAD
			continue
		if re.match(r"^.+-[0-9]+$", name) == None:
			continue
		basename = name[:name.rfind("-")]
		suffix = name[name.rfind("-"):]

		for sp in nets:
			if sp.tag != "net":
				continue
			name2 = sp.attrib["name"]
			if re.match(r"^Net-\([A-Za-z0-9\-]+\)", name2) != None:
				continue
			if re.match(r"^.+-[0-9]+$", name2) == None:
				continue
			basename2 = name2[:name2.rfind("-")]
			# ペアになるnetを見つけた。既にペア登録があれば追加、でなければ新規生成する。
			if basename == basename2:	# 末尾の数字以外の部分が一致した
				if name != name2:	# 自分自身?
					setPair(np, name2)
					setPair(sp, name)
	return(None)
#	for np in nets:
#		if np.tag == "net":
#			if np.get("pair", default=None) != None:
#				continue	# already paired
#			name = np.attrib["name"]
#			if re.match(r"^Net-\([A-Za-z0-9\-]+\)", name) != None:
#				continue
#			if re.match(r"^.+-[01]$", name) == None:
#				continue
#			nlen = len(name)
#			suffix = name[-1]
#			if suffix == "0":
#				pname = name[:nlen-1]+"1"
#			else:
#				pname = name[:nlen-1]+"0"
#			#print("original name: {0}, look for {1}".format(name, pname))
#			for sp in nets:
#				if sp.tag == "net":
#					if sp.attrib["name"] == pname:
#						np.set("pair", pname)
#						if sp.get("pair", default=None) == None:
#							sp.set("pair", name)
#						break
#	return(None)


#
#	getValue() ----- 与えられた引数がintまたはfloatの場合は、値をstrにして返す。単位はNone。引数が文字列だった場合は数値を表
#		す文字列＋単位（があるかも知れない）として解釈する。数値はstrにして、また単位（と思われる文字列）はそのまま返す。
#

def getValue(val):
	if type(val) in [int, float]:
		return(str(val), None)
	# In case of a string is given...
	pat = re.compile("^[-+]?([0-9]+(\.[0-9]*)?|\.[0-9]+)([eE][-+]?[0-9]+)?")
	r = re.match(pat, val)
	return(float(val[r.start():r.end()]), val[r.end():])


#
#	getValueWithUnit() ----- 文字列を数値と単位に分離し、単位がない場合にデフォの単位を補給しつつXMLの要素を生成する
#

def getValueWithUnit(vals, du, tg):
	v, u = getValue(vals)
	if u == None or u == "":
		u = du
	newElem = etree.Element(tg)
	newElem.text = str(v)
	if u != None:
		newElem.set("unit", u)
	return(newElem)


#
#	netInfo() ----- <net>要素のname属性にファイバ情報が与えられている場合に、それを取り出して独立した要素に展開する
#

def netInfo(nets):
	for np in nets:
		if np.tag != "net":
			continue
		if not "name" in np.attrib.keys():
			continue
		name = np.attrib["name"]
		if re.match(r"^Net-\([A-Za-z0-9\-]+\)", name) != None:	# 自動生成の名前ならスルー
			continue
		if re.match(r"^.*{.*}", name) == None:	# JSONらしいモノが含まれてなければスキップ
			continue
		# KiCADが頭にシート名を付けてくるので除去
		aname = re.sub("^.*{", "{", name)
		try:
			aroot = json.loads(aname)
		except:
			msglog("Formatting error in fiber information of "+aname, 2)
		# シート名は名前のプリフィクスにする
		p = re.match(r"^.*{", name)
		if p:
			prefix = p.group()[0:-1]
		else:
			prefix = ""
		# json→辞書変換して順番に処理
		for ap in aroot:
			#print(ap+":"+str(aroot[ap]))
			if ap == "name":
				if type(aroot[ap]) != str:
					msglog("string expected for "+ap+":"+str(aroot[ap]), 2)
				np.attrib["name"] = prefix+aroot[ap]
			elif ap == "type":
				if type(aroot[ap]) != str:
					msglog("string expected for "+ap+":"+str(aroot[ap]), 2)
				if not aroot[ap] in ["SMF", "MMF", "DCF", "PMF"]:
					msglog("unknown fiber type "+aroot[ap], 2)
				newElem = etree.Element("type")
				newElem.text = aroot[ap]
				np.append(newElem)
			elif ap == "length":
				np.append(getValueWithUnit(aroot[ap], "km", ap))
			elif ap == "loss":
				np.append(getValueWithUnit(aroot[ap], "dB", ap))
			elif ap == "unitLoss":
				np.append(getValueWithUnit(aroot[ap], "dB/km", ap))
			elif ap == "dispersion":
				np.append(getValueWithUnit(aroot[ap], "ps/nm", ap))
			elif ap == "unitDispersion":
				np.append(getValueWithUnit(aroot[ap], "ps/nm/km", ap))
			elif ap == "cost":
				np.append(getValueWithUnit(aroot[ap], None, ap))
			elif ap == "ductList":
				if type(aroot[ap]) != list:
					msglog("list of strings expected for "+ap+":"+str(aroot[ap]), 2)
				dlst = ""
				for lp in aroot[ap]:
					if dlst != "":
						dlst += ","
					dlst += lp
				dp = np.find("ductList")
				if dp == None:
					dp = etree.Element("ductList")
					dp.text = dlst
					np.append(dp)
				else:
					dp.text = dlst
			elif ap == "ductListName":
				if type(aroot[ap]) != str:
					msglog("string expected for "+ap+":"+str(aroot[ap]), 2)
				dp = np.find("ductList")
				if dp == None:
					dp = etree.Element("ductList")
					dp.set("ductListName", aroot[ap])
					np.append(dp)
				else:
					dp.set("ductListName", aroot[ap])
			else:
				msglog("unknown fiber attribute: "+ap, 2)


#
#	lossAndDispersion() ----- 損失と分散の属性値がないファイバについて、長さとファイバの種類が指定
#		されていれば、単位長当たりの値から計算で求めた値を補填する
#		ついでに損失と分散の最大値を求める。

def lossAndDispersion(nets):
	global	costEquation, unitLoss, unitDispersion
	global	maxLoss, maxDispersion

	floss = costEquation.find("%l") >= 0 or costEquation.find("%L") >= 0
	fdisp = costEquation.find("%d") >= 0 or costEquation.find("%D") >= 0 
	for np in nets:
		if np.tag != "net":
			continue
		lp = np.find("length")
		tp = np.find("type")

		# 損失についての処理
		ap = np.find("loss")
		if ap != None:
			loss = float(ap.text)
		else:
			loss = None
		if floss and np.find("loss") == None:	# <cost>の計算に損失が必要なのに値が指定されていない時
			if lp != None:		# 長さの指定は必須→単位長当たりの損失の値があっても長さがわからなければ計算のしようがない
				if np.find("unitLoss") != None:
					loss = float(lp) * float(np.find("unitLoss"))
			# 長さがわかって単位長当たりの損失の値がない→デフォの値を使って計算（ファイバの種類が必要）
				elif tp != None:
					if tp in unitLoss:
						loss = float(lp) * float(unitLoss[tp])
			if loss != None:	# 計算で求めた損失値を属性に追加
				ls = etree.Element("loss")
				ls.set("unit", "dB")
				ls.text = str(loss)
				np.append(ls)
		if loss != None:
			if loss > maxLoss:
				maxLoss = loss

		# 分散についての処理
		dp = np.find("dispersion")
		if dp != None:
			disp = float(dp.text)
		else:
			disp = None
		if fdisp and np.find("dispersion") == None:	# <cost>の計算に分散が必要なのに値が指定されていない時
			if lp != None:		# 長さの指定は必須→単位長当たりの分散の値があっても長さがわからなければ計算のしようがない
				if np.find("unitDispersion") != None:
					disp = float(lp) * float(np.find("unitDispersion"))
			# 長さがわかって単位長当たりの分散の値がない→デフォの値を使って計算（ファイバの種類が必要）
				elif tp != None:
					if tp in unitDispersion:
						disp = float(lp) * float(unitDispersion[tp])
			if disp != None:	# 計算で求めた分散値を属性に追加
				ds = etree.Element("dispersion")
				ds.set("unit", "ps/nm")
				ds.text = str(disp)
				np.append(ds)
		if disp != None:
			if disp > maxDispersion:
				maxDispersion = disp


#
#	netCosts() ----- <net>に与えられた属性の値から接続コストを計算して新しい属性<cost>として追加する。
#		<cost>の値は0.0～1.0の範囲で、値が小さいほど接続コストが小さい==優先的に選択されるべき。
#		計算はAdditinalInfoファイルで与えられた計算式に属性値を当てはめて算出する。属性値が欠けて
#		いる場合は<cost>の値はすべて0.0となる。
#		本バージョンでは<cost>の計算に必要な属性は損失と分散及びそれぞれの最大値に限定する。
#

def netCosts(nets):
	global	costEquation
	global	maxLoss, maxDispersion

	floss = costEquation.find("%l") >= 0 or costEquation.find("%L") >= 0
	fdisp = costEquation.find("%d") >= 0 or costEquation.find("%D") >= 0 
	for np in nets:
		if np.tag != "net":
			continue
		if np.find("cost") != None:	# it already has a <cost>
			continue
		lp = np.find("loss")
		dp = np.find("dispersion")
		if (floss and lp == None) or (fdisp and dp == None):
			cost = 0.0
		else:
			if lp != None:
				loss = float(lp.text)
			if dp != None:
				disp = float(dp.text)
			eq = costEquation
			while eq.find("%") >= 0:
				eq = re.sub("%L", str(maxLoss), eq)
				eq = re.sub("%D", str(maxDispersion), eq)
				eq = re.sub("%l", str(loss), eq)
				eq = re.sub("%d", str(disp), eq)
			#print(costEquation+" -> "+eq)
			try:
				cost = round(eval(eq), 3)
			except:
				msglog("error in cost equation: "+costEquation, 2)
		#print(cost)
		cs = etree.Element("cost")
		cs.text = str(cost)
		np.append(cs)


#
#	pinMatch() ----- ピン番号の一致をみる。plstrは文字列で、「範囲と列挙」のリストである。
#

def pinMatch(n, plstr):
	plist = plstr.split(',')
	for pp in plist:
		if pp.isdigit():	# 単一の番号
			if int(pp) == n:
				return True
		else:	# p1-p2 ?
			if pp.find('-') > 0:	# p1-p2
				rng = pp.split('-')
				if len(rng) == 2:
					p1 = rng[0]
					p2 = rng[1]
					if p1.isdigit() and p2.isdigit():
						p1n = int(p1)
						p2n = int(p2)
						if p1n < p2n:
							if p1n <= n and n <= p2n:
								return True
						elif p2n <= n and n <= p1n:
							return True
	return False


#
#	findPinAttr() ----- 与えられたピン番号にマッチする属性を探す。
#		ピン番号はintで与えられる。属性の方はjsonをロードした辞書であり、キー"pin"の値は：
#		(1) 1つのピン番号を表す整数文字列
#		(2) 範囲と列挙（p1-p2, p3, p4）
#		(3) *(任意の番号にマッチ)
#		のいずれかである。JSON書式内では順番が保存されないのでいきなり(3)が出てくると無条件にマッチしてしまう。
#		そのため、先ず(3)以外でマッチするかどうかを試し、一致しなければ(3)を探す。
#

def findPinAttr(num, attr):
	if not "pinAttribute" in attr:
		msglog("pinAttribute not found", 2)
	for ap in attr['pinAttribute']:
		if 'pin' in ap:
			if ap['pin'] == '*':
				continue
			if pinMatch(num, ap['pin']):
				newattr = ap.copy()
				newattr.pop('pin', None)	# remove 'pin'
				return(newattr)
	# 「範囲と列挙」に一致しなかったので*を探す
	for ap in attr['pinAttribute']:
		if 'pin' in ap:
			if ap['pin'] == '*':
				newattr = ap.copy()
				newattr.pop('pin', None)	# remove 'pin'
				return(newattr)
	return(None)


#
#	components() ----- This function converts <compnents> part og the source XML into "devices" and "pseudodevices".
#

def components(comp, libs, nets):
	global	devices, pseudodevices

	totalerr = 0
	for item in comp:
		if item.tag != 'comp':
			continue
		# 一つの<comp>エントリの処理はここから始まる
		errcnt = 0
		# 参照番号(N101のような)を得る。見つからなければソースXMLが壊れているので続行しない。
		refnum = None
		try:
			refnum = item.attrib['ref']	# reference number
		except:
			pass
		if refnum == None:
			errcnt = errcnt + 1
			msglog("reference number not seen.", 2)

		# 次にデバイスのクラス名について同様の処理を行う。
		if item.find("value") != None:
			classname = item.find("value").text
		else:
			errcnt = errcnt + 1
			msglog("device class name not found.", 2)

		# sheetpathの先頭部分は帰属先ノード名として扱う。
		try:
			sheetpath = item.find("sheetpath").attrib['names']
		except:
			errcnt = errcnt + 1
			msglog("sheetpath name not found.", 2)
#		if "ODU" in sheetpath:
#			supernetwork = "ODUNW"
#		else:
#			supernetwork = "LAMBDANW"

		# このデバイスのライブラリに関する情報を得る。ライブラリ名とデバイス名。
		# ライブラリ情報がない、或いは見つからない場合はソースXMLが壊れている。
		if item.find("libsource") != None:
			libsource = item.find("libsource")
			try:
				libpart = libsource.attrib['part']
				libname = libsource.attrib['lib']
			except:
				errcnt = errcnt + 1
				msglog("library information is defected.", 2)
		else:
			errcnt = errcnt + 1
			msglog("comp losts libsource", 2)

		# ユーザ定義フィールドを拾う。知ってるフィールドは出力へ、知らない奴が居たらエラー
		# フィールドが全くない場合はソースXMLが壊れている。
		flds = item.find("fields")
		if flds == None:
			errcnt = errcnt + 1
			msglog("no user field in comp. "+refnum, 2)
		else:	# userFieldsに指定されたフィールドを拾う。pinAttributeだけは後で別の処理に使うので特別扱い。
			defindField = {}
			fldPinAttr = None
			for ff in flds:
				if ff.tag == 'field':
					if "name" not in ff.attrib:
						errcnt = errcnt + 1
						msglog("name attribure is lost in field")
					elif ff.attrib['name'] in userFields:
						defindField[ff.attrib['name']] = ff.text
						if ff.attrib['name'] == "pinAttribute":
							try:
								fldPinAttr = json.loads(ff.text)
							except:
								msglog("Formatting error in PinAttribute of "+refnum+" "+ff.text, 2)
					else:
						errcnt = errcnt + 1
						msglog("unexpected user field "+ff.attrib['name']+" in "+refnum, 2)

		nid = "/TEST_" #urnprifix+sheetpath.replace('/', '_')
		idname = nid+classname+"_"+refnum #fldID

# 今次バージョンではphyAttribytes.jsonからではなくPinAttrフィールドからピン属性を得るようになるので、papは要らない。
#		pap = None
#		try:
#			pap = padict[fldModel]#classname]
#		except:
#			pass

		# ライブラリ名と部品名を手掛かりにライブラリ内を探す。
		# ライブラリにはピン番号とピンの名称、入出力の別などが記載されている。一方PinAttrフィールドにはプロトコルや波長チャネルが
		# 指定されている。両者をマージする。
		newpins = {}	# マージされたピン情報の置き場所
		for lp in libs:		# libs はソースXMLの<libparts>～</libparts>を読み込んだ辞書
			if lp.tag != 'libpart':
				continue
			if lp.attrib['part'] != libpart or lp.attrib['lib'] != libname:
				continue
			# ライブラリ内に<pins>がなければソースXMLが壊れている
			ps = lp.find('pins')
			if ps == None:
				msglog("<pins> not found in "+libpart, 2)
			# ライブラリにあるピン情報を1つずつ取り出してPinAttrとマージしてnewpinsに入れる
			for pp in ps:
				if pp.tag != 'pin':
					continue
				# ライブラリのピン情報には番号、名前、I/Oの区別、が期待される。
				try:
					pinnum = int(pp.attrib['num'])
					pinlbl = pp.attrib['name']
					pintyp = pp.attrib['type']
				except:
					msglog("Lack of pin information in "+libpart, 2)
				# 制御対象デバイス(参照番号がNで始まる)でない場合はPinAttrがなくてもよい
				if refnum[0] != 'N' and fldPinAttr == None:
					newattr = {}
					newattr['name'] = idname+'_'+pp.attrib['name']
					newattr["io"] = pintyp
					newpins[pp.attrib['num']] = newattr

				# PinAttrに与えられた属性から、pinnumにマッチするものを選び出す。
				else:
					if fldPinAttr == None:
						msglog("Pin attribute not found in "+refnum, 2)
					newattr = findPinAttr(pinnum, fldPinAttr)
					if newattr == None:
						msglog("pin attribute not found in "+refnum, 2)
					newattr['name'] = idname+'_'+pp.attrib['name']
					newattr["io"] = pintyp
					if refnum[0] == 'N':
						exta = extraAttr(refnum, pinnum, nets, comp)
						if exta != None:
							if 'Type' in exta:
								exta.pop('Type')
							newattr.update(exta)
					newpins[pp.attrib['num']] = newattr

			if newpins == {}:
				msglog("port information not found in "+libpart, 2)

			# Is a new device is unique?
			if refnum in devices:
				errcnt = errcnt + 1
				msglog("Duplicated reference number: "+idname)
			for dp in devices.values():
				if idname in dp.values():
					errcnt = errcnt + 1
					msglog("Duplicated device name: "+idname)

			# if new device is converted successfully, put it into devices[] otherwise report error.
			if errcnt > 0:
				errplace = etree.tostring(item)
				msglog("Above error(s) are found in following part of input XML:")
				msglog(errplace.decode(sys.stdin.encoding))
				totalerr = totalerr + 1
			else:
				if "Type" in defindField:
					fldType = defindField["Type"]
				else:
					msglog("Type is missing in "+libpart, 2)
				#newcomp = {"idname" : idname, "type" : fldType, "residence" : sheetpath}
				newcomp = {"idname" : idname, "residence" : sheetpath}
				for finx in defindField:
					if finx != "pinAttribute":
						newcomp[finx] = defindField[finx]
				if len(newpins) > 0:
					newcomp['ports'] = newpins
				if fldType == "PORTATTR":
					pseudodevices[refnum] = newcomp
				else:
					devices[refnum] = newcomp



#
#	getAddInfo() ----- チャネル・テーブルなど、付加的な情報が入ったファイルを読む。
#		<urnprifix>と<defaultCostInfo>は本プログラム内で参照するので内部表現に変換する。
#		<channelInfo>と<signalformInfo>はスルーしてそのまま出力する。
#

def getAddInfo(name):
	global	urnprifix, costEquation, unitLoss, unitDispersion, channelInfo, signalformInfo

	try:
		tree = etree.parse(name)
	except:
		return("Error in parsing: "+name)

	root = tree.getroot()

	# <urnPrifix>
	pfx = root.find("urnPrifix")
	if pfx == None:
		return("<urnPrifix> not found")
	urnprifix = pfx.text

	# <defaultCostInfo>
	cinfo = root.find("defaultCostInfo")
	if cinfo == None:
		return("<defaultCostInfo> not found")
	# <defaultCostInfo>に含まれるのは<uitLoss>と<unitDispersion>と<costEquation>の3種類の要素のみ
	isFloat_r = re.compile('^[+-]?(\d*\.\d+|\d+\.?\d*)([eE][+-]?\d+|)\Z')
	for cp in cinfo:
		if cp.tag == "unitLoss":
			if not "fiberType" in cp.attrib:
				return("fiberType not seen in <unitLoss>")
			ft = cp.attrib["fiberType"]
			if ft in unitLoss:
				return("duplicated unitLoss: "+ft)
			if not isFloat_r.match(cp.text):
				return("numeric expected: "+cp.text)
			unitLoss[ft] = float(cp.text)
		elif cp.tag == "unitDispersion":
			if not "fiberType" in cp.attrib:
				return("fiberType not seen in <unitDispersion>")
			ft = cp.attrib["fiberType"]
			if ft in unitDispersion:
				return("duplicated unitDispersion: "+ft)
			if not isFloat_r.match(cp.text):
				return("numeric expected: "+cp.text)
			unitDispersion[ft] = float(cp.text)
		elif cp.tag == "costEquation":
			if costEquation != None:
				return("<constEquation> duplicated")
			costEquation = cp.text
		else:
			return("unexpected tag in <defaultCostInfo> "+cp.tag)
	#print(costEquation)
	#print(unitLoss)
	#print(unitDispersion)

	# <channelInfo>, <signalformInfo>を読み込んでおく（後でそのまま出力ファイルへ
	channelInfo = root.find("channelInfo")
	if channelInfo == None:
		return("<channelInfo> not found")
	signalformInfo = root.find("signalformInfo")
	if signalformInfo == None:
		return("<signalformInfo> not found")

	return(None)	# successful


#
#	reconEtree() ----- 書式を整えるために再構築する
#

def reconEtree(tp):
	newElem = etree.Element(tp.tag)
	for ap in tp.attrib:
		newElem.set(ap, tp.attrib[ap])
		newElem.text = tp.text
	for ep in tp:
		newElem.append(reconEtree(ep))
	return(newElem)


def main():
	'''Reads specified XML file, and then convert it into the Topology file.
	python xmp2topology2 -a additionalImfo.json -o output-file [-v] input-file
	'''
	# KiCADのネットリスト生成用プラグインを起動するロジックのバグで入出力ファイルの名前が同じになってしまう。
	# これを回避するため、出力ファイル名は渡された入力ファイル名（変換前のXMLが入っている）のベース部分に
	# _tplを付けたものにする。KiCADでの起動コマンド行はpython "xml2topoloy2.py" -a "AddtionalInfo,xml" 入力ファイル名
	# となる。ログファイルは実行ファイルと同じ場所に決め打ちの名前で作る。
	global lgfd, debug, comp, libs, nets, urnprifix, channelInfo, signalformInfo

	try:
		oplist, args = getopt.getopt(sys.argv[1:], 'dva:l', ['debug', 'verbose', 'additionalinfo', 'log'])
	except getopt.GetoptError:
		sys.exit(2)

	# Tk will be used to pop an error message up.
	tkroot = tkinter.Tk()
	tkroot.withdraw()

	ofname = None		# output file name
	paname = None		# attribute file name
	lgname = None		# log file name
	verbose = False
	debug = False
	for opt, arg in oplist:
		if opt == '-d' :
			debug = True
			print("debug")
		if opt == '-v' :
			verbose = True
		if opt == '-a' :
			paname = arg
		if opt == '-l' :
			lgname = os.path.split(sys.argv[0])[0]+"\\xml2topoloy2.log"
	ifname = args[0]

	# AdditionalInfoファイル名のデフォルトは、実行ファイルと同じ場所のAdditionalInfo.xml
	if paname == None:
		paname = os.path.split(sys.argv[0])[0]+"\\AdditionalInfo.xml"
	# 出力ファイル名は入力ファイル名から自動的に生成する。
	ofname = os.path.splitext(ifname)[0]+"_tpl.xml"

	if debug:
		print(verbose)
		print(paname)
		print(ofname)
		print(ifname)
		print(lgname)
	
	# open Log file if necessary
	if lgname == None:
		lgfd = None
	else:
		try:
			lgfd = open(lgname, 'w')
		except:
			tkmsg.showerror("physch2xml", "Log file "+lgname+" open error. further execution aborted.")
			sys.exit(2)
		lgfd.write("xml2topology started at "+time.asctime()+"\n")
		lgfd.write(ifname+"\n")
		lgfd.write(ofname+"\n")
		lgfd.write(paname+"\n")

	# read AdditionalInfo.xml
	r = getAddInfo(paname)
	if r != None:
		msglog(r, 2)

	# getting a source XML into a tree
	try:
		tree = etree.parse(ifname)
	except:
		msglog("Error in parsing: "+ifname, 3)
	root = tree.getroot()

	# read port/net attribute file if specified. It also includes urn prifix.
	if urnprifix == None:
		urnprifix = "urn:pathfinder2_test:"
#	if paname != None:
#		try:
#			paf = open(paname)
#			AttrDict = json.load(paf)
#		except:
#			tkmsg.showerror("physch2xml", "Error in parsing: "+paname)
#			msglog("Error in reading " + paname, 3)
#		if 'urlprifix' in AttrDict:
#			urnprifix = AttrDict['urnprifix']
#		if 'ConstraintNames' in AttrDict:
#			ConstraintNameList = AttrDict['ConstraintNames']
#		else:
#			ConstraintNameList = []
#		padict = AttrDict["PinAttribute"]
#		if 'NetAttribute' in AttrDict:
#			nadict = AttrDict["NetAttribute"]
#		else:
#			nadict = []

	# (1) <design> element will be converted
	dsgn = root.find("design")
	if dsgn == None:
		msglog("<design> not found", 3)

	design_info = {}	# JSON's object is Python's dictionary
	for item in dsgn:
		if item.tag == 'source' :
			design_info['schematics'] = item.text
		elif item.tag == 'date' :
			design_info['date'] = item.text

	# before reading other parts, "nets" should be obtained because it will be scanned to find extra-attribute for the ports.
	nets = root.find('nets')
	if nets == None:
		msglog("<nets> is lost", 3)

	# (2) <components> and <libparts> are read and converted into device-list
	comp = root.find("components")
	if comp == None:
		msglog("<components> not seen.", 3)

	libs = root.find("libparts")
	if libs == None:
		msglog("<libparts> not seen.", 3)

	components(comp, libs, nets)

#				if 'Model' in newcomp:
#					if 'GLPK' in newcomp:
#						print(newcomp['Model'], newcomp['GLPK'])
#					else:
#						print("no GLPK in {0}".format(newcomp['Model']))

	#print(devices)
	#print("len(nets) = {0}".format(len(nets)))
	# nodeが1個のnet（行き止まり）は削除
	# リストの要素をforしながらいきなり削除するとおかしくなるので、1個消したら最初からやり直す
	while True:
		dcnt = 0
		for np in nets:
			if np.tag == "net":
				if len(np) < 2:
					nets.remove(np)
					dcnt += 1
					break;
				elif len(np) > 2:
					msglog("more than two nodes in net {0}".format(np.attrib["code"]), 1)
		if dcnt == 0:
			break;
	#print("len(nets) = {0}".format(len(nets)))


	# pseudodevicesにあるデバイスの2つのポートにつながるネットを接続する
	# 最大2連続の仮想デバイスを処理するため、処理対象が無くなるまで繰り返す
	while True:
		#print("len(nets) = {0}".format(len(nets)))
		dcnt = 0
		for ref in pseudodevices:
			dev = pseudodevices[ref]
			n1 = findNetByRefPin(nets, ref, "1")
			n2 = findNetByRefPin(nets, ref, "2")
			if n1 == None or n2 == None:
				continue	# when it does not has two ports, do nothing
			if len(n1) != 2 or len(n2) != 2:
				continue	# dead end net
			c1 = isConnectToPseudoDevice(comp, n1, ref, "1")
			c2 = isConnectToPseudoDevice(comp, n2, ref, "2")
			# 3連続の仮想デバイスはあり得ないので、どちらかが実デバイスにつながっているはず
			# 仮想デバイスの両側が実デバイスと言う状況（界面にSTPが１つだけ）は変だがあり得る
			if c1 and c2:
				msglog("device {0} connected to pseudo-device in both sides: {1}, {2}".format(ref, n1.attrib["code"], n2.attrib["code"]), 1)
			# 名前を引き継ぐ。Net-で始まる名前は自動で付与されたものなので、そうでない名前（ユーザ指定）を優先する
			m1 = n1.attrib["name"]
			m2 = n2.attrib["name"]
			if re.match(r"^Net-\([A-Za-z0-9\-]+\)", m1) == None:
				m0 = m1
			else:
				m0 = m2
			# 注目している仮想デバイスの、実デバイスにつながっている側のネットの、仮想デバイス側を、
			# 反対側のネットの仮想デバイスではない側で置き換える。
			replaced = False
			# real <a1---n1---b1> pseudodevices[ref] <a2---n2---b2> pseudo である時、b1 = b2 とする
			if not c1:	# n1側が実デバイス
				b2 = getFarSide(n2, ref, "2")
				if b2 == None:
					msglog("Far side not found")
				for np in nets:
					if np.tag == "net":
						for sp in np:
							if sp.tag == "node":
								if ref == sp.attrib["ref"] and "1" == sp.attrib["pin"]:
									#print(etree.tostring(np, "utf-8"))
									np.remove(sp)
									np.append(b2)
									#print(etree.tostring(np, "utf-8"))
									markRemove(nets, n2)#nets.remove(n2)
									np.attrib["name"] = m0
									replaced = True
									break
						if replaced:
							break
			else:	# pseudo <a1---n1---b1> pseudodevices[ref] <a2---n2---b2> real である時、a2 = a1 とする
				a1 = getFarSide(n1, ref, "1")
				if a1 == None:
					msglog("Far side not found")
				for np in nets:
					if np.tag == "net":
						for sp in np:
							if sp.tag == "node":
								if ref == sp.attrib["ref"] and "2" == sp.attrib["pin"]:
									#print(etree.tostring(np, "utf-8"))
									np.remove(sp)
									np.append(a1)
									#print(etree.tostring(np, "utf-8"))
									markRemove(nets, n1)#nets.remove(n1)
									np.attrib["name"] = m0
									replaced = True
									break
						if replaced:
							break
			if replaced:
				dcnt += 1
			else:
				msglog("combining net {0} with {1} is failed".format(n1.attrib["code"], n2.attrib["code"]), 1)
			if dcnt == 0:
				break
		#print(dcnt)
		if dcnt == 0:
			break
		#print("len(nets) = {0}".format(len(nets)))

	# remove tag="removed" in nets
	while True:
		dcnt = 0
		for np in nets:
			if np.tag == "removed":
				nets.remove(np)
				dcnt += 1
				break;
		if dcnt == 0:
			break;

	# <net>要素内のname属性に含まれている（かも知れない）ファイバの属性情報を取り出して独立した要素に展開
	netInfo(nets)
	# 損失と分散がないファイバについて、長さとファイバの種類が指定されていれば、単位長当たりの値から計算で求めた値を補填する
	# 損失と分散の最大値を求める。
	lossAndDispersion(nets)
	# すべてのファイバについて、損失や分散の値があればcostを計算する。なければcostは0
	netCosts(nets)

	# 名前を手掛かりにネットのペアリングを行う
	pairPath(nets)

	# netsのままtopologyにappendすると書式が乱れるので再構築する
	newNets = etree.Element("nets")
	for np in nets:
		if np.tag == "net":
			nnet = etree.Element("net")
			nnet.set("code", np.attrib["code"])
			nnet.set("name", np.attrib["name"])
			if np.get("pair", default=None) != None:
				nnet.set("pair", np.attrib["pair"])
			for op in np:
				if op.tag == "node":
					nnode = etree.Element("node")
					nnode.set("ref", op.attrib["ref"])
					nnode.set("pin", op.attrib["pin"])
					nnet.append(nnode)
			newNets.append(nnet)

	#print("len(newNets) = {0}".format(len(newNets)))

	# convert internal representations of devices into XML
	comps = componentXml(devices)
	#print(miniDom.parseString(etree.tostring(comps, "utf-8")).toprettyxml(indent=" "))

	# designについてもnets同様書式を整えるために再構築する
	newDesign = etree.Element("design")
	for dp in dsgn:
		if dp.tag == "source" or dp.tag == "date" or dp.tag == "tool":
			ds = etree.Element(dp.tag)
			ds.text = dp.text
			newDesign.append(ds)
		elif dp.tag == "sheet":
			ds = etree.Element("sheet")
			for ap in ["number", "name", "tstamps"]:
				ds.set(ap, dp.attrib[ap])
			for pp in dp:
				if pp.tag == "title_block":
					ts = etree.Element("title_block")
					for tp in pp:
						tm = etree.Element(tp.tag)
						tm.text = tp.text
						if tp.tag == "comment":
							for ap in ["number", "value"]:
								tm.set(ap, tp.attrib[ap])
						ts.append(tm)
					ds.append(ts)
			newDesign.append(ds)
	newDesign.append(reconEtree(channelInfo))
	newDesign.append(reconEtree(signalformInfo))

	topology = etree.Element("topology")
	topology.append(newDesign)
	topology.append(comps)
#	topology.append(newNets)
#	topology.append(reconEtree(nets))
	topology.append(nets)

	topologyXml = miniDom.parseString(etree.tostring(topology, "utf-8")).toprettyxml(indent="    ")
#	topologyXmla = re.sub(r"^ *\r\n", "X", topologyXml)
	noemp = re.compile("^[ \t]*\n", re.MULTILINE)
	topologyXmla = re.sub(noemp, "", topologyXml)
	#print(topologyXmla)

	# convert internal representations into XML
	#jj = json.dumps({"design_info" : design_info, "constraints" : ConstraintNameList, "devices" : devices}, sort_keys=True, indent=4)

	if ofname != None:
		of = open(ofname, mode="wt", encoding="utf-8")
		print(topologyXmla, file=of)
		of.close()
		if verbose:
			tkmsg.showinfo("physch2xml", "Conversion completed successfully. Output file is: "+ofname)

	# close log-file if necessary
	if lgfd != None:
		lgfd.write("physch2json finished successfully at "+time.asctime()+"\n")
		lgfd.close()

if __name__ == '__main__' :
	main()
